searchNodes=[{"type":"module","title":"Q","doc":"Q Documentation for Q .","ref":"Q.html"},{"type":"function","title":"Q.exec/1","doc":"Executes the queue, running all operations in the order they were added. Each operation is provided with the state of the queue as of the start of the operation. If an operation function returns an error, the execution of the queue is halted and an error is returned. Example iex&gt; Q . new ( ) ...&gt; |&gt; Q . run ( :write , fn _ -&gt; { :ok , &quot;Hello world!&quot; } end ) ...&gt; |&gt; Q . exec ( ) { :ok , %{ write : &quot;Hello world!&quot; } }","ref":"Q.html#exec/1"},{"type":"function","title":"Q.inspect/2","doc":"Adds an operation to the queue that, when executed, prints the current state of the queue. The printing is performed using Elixir's IO.inspect/2 , which provides a human-readable representation of the queue's state. Params queue : The queue to inspect. opts : Options passed to IO.inspect/2 . If the :only option is provided, only those keys will be printed.","ref":"Q.html#inspect/2"},{"type":"function","title":"Q.new/0","doc":"Creates and returns a new, empty Q struct. The resulting queue can be used as a starting point for queue operations, such as put/3 and run/4 . Example iex&gt; Q . new ( ) |&gt; Q . to_list ( ) [ ]","ref":"Q.html#new/0"},{"type":"function","title":"Q.operations/1","doc":"Returns a list of operation names stored in the queue, in the order they were added. This function can be used to review the sequence of operations added to the queue. Example iex&gt; Q . new ( ) ...&gt; |&gt; Q . put ( :init , &quot;foo&quot; ) ...&gt; |&gt; Q . run ( :test , fn text -&gt; { :ok , text } end , [ :init ] ) ...&gt; |&gt; Q . operations ( ) [ &quot;test&quot; , &quot;init&quot; ]","ref":"Q.html#operations/1"},{"type":"function","title":"Q.put/3","doc":"Adds a value to the state so far under the given name. This function modifies the queue by associating the given name with a value . If the operation is successful, the name will become a key in the queue's internal state map, and value will be its associated value. Params queue : The queue to add the value to. name : The name to associate with the value. It must be unique, or an error will be raised. value : The value to add to the queue. Errors This function will raise an error if name is already present in the queue. Example iex&gt; Q . new ( ) ...&gt; |&gt; Q . put ( :params , %{ foo : &quot;bar&quot; } ) ...&gt; |&gt; Q . exec ( ) { :ok , %{ params : %{ foo : &quot;bar&quot; } } }","ref":"Q.html#put/3"},{"type":"macro","title":"Q.queue/1","doc":"Provides a DSL for creating a queue using a more declarative syntax. Example queue do put ( :input , &quot;some data&quot; ) run ( :processed , &amp; process_data / 1 , [ :input ] ) exec ( ) # optional, will return the result immediately end","ref":"Q.html#queue/1"},{"type":"function","title":"Q.run/4","doc":"Adds a function to be executed as part of the queue. This function enables adding operations to the queue that are executed when exec/1 is called. The operations can be anonymous functions or functions in a module. The function should return either {:ok, value} if the operation was successful or {:error, value} if the operation failed. The state of the queue is passed as arguments to the function if no params are specified. Functions can also return {:halt, value} to halt execution at any point. Params queue : The queue to which the function should be added. name : A unique name for this operation. It will be used as a key in the queue's internal state. fun : The function to be run. It can be an anonymous function or a tuple containing a module, function, and arguments. params (optional): A list of keys in the queue's internal state that should be passed to the function as arguments.","ref":"Q.html#run/4"},{"type":"function","title":"Q.run/6","doc":"Adds a function to be executed as part of the queue. Similar to run/4 , but allows to pass module name, function and arguments. The function should return either {:ok, value} if the operation was successful or {:error, value} if the operation failed. The state of the queue is passed as arguments to the function if no params are specified. Functions can also return {:halt, value} to halt execution at any point. Params queue : The queue to which the function should be added. name : A unique name for this operation. It will be used as a key in the queue's internal state. mod : The module where the function is defined. fun : The function to be executed. args : Arguments that should be passed to the function. params (optional): A list of keys in the queue's internal state that should be passed to the function as arguments.","ref":"Q.html#run/6"},{"type":"function","title":"Q.to_list/1","doc":"Converts the operations queue to a list, in the order they were added. This function is useful for inspecting the current state of the queue. Direct inspection of the queue's internal state is discouraged to maintain encapsulation.","ref":"Q.html#to_list/1"},{"type":"type","title":"Q.exec_error/0","doc":"","ref":"Q.html#t:exec_error/0"},{"type":"type","title":"Q.fun_arity1/0","doc":"","ref":"Q.html#t:fun_arity1/0"},{"type":"type","title":"Q.fun_mfa/0","doc":"","ref":"Q.html#t:fun_mfa/0"},{"type":"type","title":"Q.name/0","doc":"","ref":"Q.html#t:name/0"},{"type":"type","title":"Q.state/0","doc":"","ref":"Q.html#t:state/0"},{"type":"type","title":"Q.t/0","doc":"","ref":"Q.html#t:t/0"},{"type":"extras","title":"Overview","doc":"Q Q is a tool for building and executing pipelines of operations, influenced by Ecto.Multi. It preserves all intermediate results so that you retain clarity and visibility into your data transformations at each step. When a pipeline fails, you can see precisely which step broke and what the state looked like at that point.","ref":"readme.html"},{"type":"extras","title":"Overview - Why Q?","doc":"Q simplifies building pipelines in three key ways: Never lose context : Q like Ecto.Multi will always return all executed steps results. When a step breaks, Q returns all prior step results along with the failing step’s error. Control passed parameters : Steps can operate on the entire pipeline state or on specific labels only. This reduces the need to rewrite functions to suite the pipeline or having to introduce anonyomous functions just to be able to pick the parameters. Flexible Composition : Q queues can be composed just like normal Elixir functions. Smaller pipelines can be combined to form larger ones.","ref":"readme.html#why-q"},{"type":"extras","title":"Overview - Example","doc":"A pipeline (or “queue”) is a collection of labeled steps that transform data. Each step can take the entire pipeline state or selected portions of it. Results are always labeled and retained: defmodule UploadProcessor do import Q def process_upload ( file_data ) do queue do put ( :file_data , file_data ) run ( :parsed , &amp; DataProcessor . parse_csv / 1 , [ :file_data ] ) run ( :validated , fn %{ parsed : rows } -&gt; Enum . reduce_while ( rows , { :ok , [ ] } , fn row , { :ok , acc } -&gt; case DataProcessor . validate_row ( row ) do { :ok , valid } -&gt; { :cont , { :ok , [ valid | acc ] } } { :error , _ } = err -&gt; { :halt , err } end end ) end ) run ( :enriched , { DataProcessor , :enrich_data , [ [ use_ai : true ] ] } , [ :validated ] ) end |&gt; exec ( ) end end","ref":"readme.html#example"},{"type":"extras","title":"Overview - Error handling","doc":"One of the major benefits of using Q is that it provides a complete context even in case of errors - you know not only which step failed but have access to all results leading up to the error. def process_data ( data ) do queue do put ( :raw_data , data ) run ( :decoded , &amp; Jason . decode / 1 , [ :raw_data ] ) run ( :validated , &amp; validate_schema / 1 , [ :decoded ] ) run ( :normalized , &amp; normalize_data / 1 , [ :validated ] ) run ( :enriched , &amp; add_metadata / 1 , [ :normalized ] ) exec ( ) end end # If JSON parsing fails { :error , :decoded , % Jason.DecodeError { data : &quot;invalid&quot; , position : 0 , token : nil } , %{ raw_data : &quot;invalid&quot; } } # If validation fails, notice both previous processing steps results are available { :error , :validated , :invalid_schema , %{ raw_data : &quot;{ \\&quot; foo \\&quot; : 123}&quot; , decoded : %{ &quot;foo&quot; =&gt; 123 } } } Without Q def process_data_without_q ( data ) do with { :ok , decoded } &lt;- Jason . decode ( data ) , { :ok , validated } &lt;- validate_data ( data ) , { :ok , transformed } &lt;- normalize_data ( validated ) , { :ok , enriched } &lt;- add_metadata ( transformed ) do { :ok , enriched } else { :error , reason } -&gt; # We know the failure reason, but not the state when it failed # as the data from of processing steps is not available here { :error , reason } end end","ref":"readme.html#error-handling"},{"type":"extras","title":"Overview - Parameter Passing","doc":"Every step function will receive the entire pipeline state by default, or you can list which keys you want to pass as function arguments. defmodule ParamDemo do import Q def run_stuff ( data ) do queue do # Store some data for subsequent steps put ( :some_data , data ) # Pass a single argument from the state run ( :single_arg , &amp; MyModule . process_single / 1 , [ :some_data ] ) # Pass the complete state run ( :custom_extract , fn %{ single_arg : val } -&gt; MyModule . do_something ( val ) end ) # Pass a multiple arguments from the state run ( :multi_arg , &amp; MyModule . process_multiple / 1 , [ :custom_extract , :some_data ] ) # The mfa version allows us to pass additional params directly # by default args will be prepended i.e. # MyModule.delete(single_arg, some_data, soft_delete: true) run ( :prepend_args , { MyModule , :delete , [ [ soft_delete : true ] ] } , [ :single_arg , :some_data ] ) # To append the arguments instead we can define the order # MyModule.changeset(%MyStruct{}, single_arg, some_data) run ( :appended_args , { MyModule , :changeset , [ % MyStruct { } ] } , { [ :single_arg , :some_data ] , order : :append } ) end |&gt; exec ( ) end end","ref":"readme.html#parameter-passing"},{"type":"extras","title":"Overview - Halting Execution Early","doc":"Sometimes you don’t need the remaining steps once a certain condition is met. Q allows functions to return {:halt, value} to end the pipeline successfully: defmodule ExpensiveProcessor do import Q def process_data ( input ) do queue do put ( :input , input ) # Check cache first run ( :cache_check , fn %{ key : key } -&gt; cached = get_from_cache ( key ) if not is_nil ( cached ) do { :halt , cached } # Skip expensive operation if cached else { :ok , :not_found } end end ) # Only runs if not in cache run ( :processed , &amp; expensive_operation / 1 , [ :input ] ) run ( :cache , &amp; cache_processed_data / 1 , [ :processed ] ) exec ( ) end end # halt early on cache hit { :ok , %{ input : %{ key : &quot;BWBeN28Vb7cMEx7Ym8AUzs&quot; , data : %{ ... } } , cache_check : %{ data : %{ ... } } } # run full pipeline when not cached { :ok , %{ input : %{ key : &quot;BWBeN28Vb7cMEx7Ym8AUzs&quot; , data : %{ ... } } , cache_check : :not_found processed : %{ data : %{ ... } } cache : :ok }","ref":"readme.html#halting-execution-early"},{"type":"extras","title":"Overview - API","doc":"Q provides two styles, the DSL and a more functional chaining API. Here are both approaches side by side. DSL import Q decode_q = queue do put ( :base64_text , &quot;aGVsbG8=&quot; ) run ( :decoded , &amp; Base . decode64 / 1 , [ :base64_text ] ) run ( :decoded_mfa , { Base , :decoded64 , [ ] } , [ :base64_text ] ) end exec ( decode_q ) # Returns: { :ok , %{ base64_text : &quot;aGVsbG8=&quot; , decoded : &quot;hello&quot; , decoded_mfa : &quot;hello&quot; } } Functional decode_q = Q . new ( ) |&gt; Q . put ( :base64_text , &quot;aGVsbG8=&quot; ) |&gt; Q . run ( :decoded , &amp; Base . decode64 / 1 , [ :base64_text ] ) |&gt; Q . run ( :decoded_mfa , { Base , :decoded64 , [ ] } , [ :base64_text ] ) Q . exec ( decode_q ) # Returns: { :ok , %{ base64_text : &quot;aGVsbG8=&quot; , decoded : &quot;hello&quot; , decoded_mfa : &quot;hello&quot; } }","ref":"readme.html#api"},{"type":"extras","title":"Overview - Installation","doc":"If available in Hex , the package can be installed by adding ex_q to your list of dependencies in mix.exs : def deps do [ { :ex_q , &quot;~&gt; 1.1&quot; } ] end Documentation can be found at https://hexdocs.pm/ex_q .","ref":"readme.html#installation"},{"type":"extras","title":"License","doc":"MIT License Copyright (c) 2024 Goran Pedić Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the &quot;Software&quot;), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. THE SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.","ref":"license.html"}]